//============================================================================
//  aPCI-8806 for PCI Linux Device Driver
//  apci8806.c
//  Author  Shoichiro Ito(ADTEK SYSTEM SCIENCE Co.,Ltd.)
//  Date    2001/03/29
//  Rev.    1.0.0
//============================================================================

#ifndef MODULE
#define MODULE
#endif // MODULE

#ifndef __KERNEL__
#define __KERNEL__
#endif // __KERNEL__

#include <linux/autoconf.h>
#if defined(CONFIG_MODVERSIONS) && !defined(MODVERSIONS)
#	define MODVERSIONS
#endif

#ifdef MODVERSIONS
#	include <linux/modversions.h>
#endif

#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/fs.h>		// used file oprations
#include <linux/pci.h>		// used pci functions
#include <asm/io.h>		// used inb, outb
#include <asm/uaccess.h>	// used verify_area
#include <asm/delay.h>		// used udelay

#include "apci8806.h"

/*
 * Split minors in two parts
 */
#define TYPE(dev)		(MINOR(dev) >> 4)	// high nibble
#define NUM(dev)		(MINOR(dev) & 0xf)	// low	nibble


#define APCI8806_NAME		"apci8806"

#define APCI8806_VENDOR_ID	0x136C		// ǥХID,٥ID פʤ
#define APCI8806_DEVICE_ID	0x8806		// ɥ饤ФϿǤʤΤ

#define PWARNING(fmt, args...)	printk(KERN_WARNING "[APCI8806]: " fmt, ## args)


/*----------------- Define Device Information ------------------------------*/

typedef struct _devinf
{
	unsigned int  io_addr;			// I/O Base Address
	unsigned long f_version;		// ե뼱̻
} devinf, *pdevinf;

static devinf	devctl[APCI8806_MAX_DEVICES];	// ǥХ
static unsigned int	uapci8806major;		// Major Device Number

static int devmajor = 0;


/*----------------- File Operations ----------------------------------------*/

static int apci8806_ioctl(struct inode*, struct file*, unsigned int, unsigned long);
static int apci8806_open(struct inode* inode, struct file* file);
static int apci8806_close(struct inode* inode, struct file* file);

static struct file_operations apci8806_fops =
{
	NULL,		// loff_t	llseek(struct file*, loff_t, int)
	NULL,		// ssize_t	read(struct file*, char*, size_t, loff_t*)
	NULL,		// ssize_t	write(struct file*, const char*, size_t, loff_t*)
	NULL,		// int		readdir(struct file*, void*, filldir_t)
	NULL,		// unsigned int poll(struct file*, struct poll_table_struct*)
	apci8806_ioctl,	// int		ioctl(struct inode*, struct file*, unsigned int, unsigned long)
	NULL,		// int		mmap(struct file*, struct vm_area_struct*)
	apci8806_open,	// int		open(struct inode*, struct file*)
	NULL,		// int		flush(struct file*)
	apci8806_close,	// int		release(struct inode*, struct file*)
	NULL,		// int		fsync(struct file*, struct dentry*)
	NULL,		// int		fasync(int, struct file*, int)
	NULL,		// int		check_media_change(kdev_t dev)
	NULL,		// int		revalidate(kdev_t dev)
	NULL,		// int		lock(struct file*, int, struct file_lock*)
};


/*----------------- Define Static Functions --------------------------------*/

static int	    init_driver_info(void);
static int	    setup_driver(void);
static void	    release_driver(void);
static int	    get_device(void);
static unsigned int get_device_base_addr(unsigned int base_addr);
static int	    regist_char_driver(void);
static int	    unregist_char_driver(void);
static int	    get_device_state(int ref);

static int	verify_arg(unsigned int cmd, unsigned long arg);

static int	apci8806_ioc_set_da_data(struct file* file, unsigned long arg);
static int	apci8806_ioc_get_exin(struct file* file, unsigned long arg);

static int	is_channel(int ch);

static inline void	    da_set_data(int ref, unsigned short data);
static inline unsigned char da_get_data(int ref);


MODULE_PARM(devmajor, "i");	// ᥸㡼ֹѥ᡼


/*==========================================================================*/
/*  init_module - ɥ饤ФϿ					    */
/*==========================================================================*/
int init_module(void)
{
	int result;

	// PCI ǽ̵ͭȽ
	if (!pci_present())
		return -ENODEV;

	// ǥХɥ饤ѿν
	result = init_driver_info();
	if (result < 0)
		return result;

	// ǥХɥ饤ФΥåȥå
	result = setup_driver();
	if (result < 0)
		return result;

	return 0;
}

/*==========================================================================*/
/*  cleanup_module - ɥ饤ФϿä				    */
/*==========================================================================*/
void cleanup_module(void)
{
	// ǥХɥ饤Фβ
	release_driver();
}

/*--------------------------------------------------------------------------*/
/*  apci8806_ioctl - ǥХͭ					    */
/*--------------------------------------------------------------------------*/
static int apci8806_ioctl(struct inode* inode, struct file* file, unsigned int cmd, unsigned long arg)
{
	int err = verify_arg(cmd, arg);
	if (err)
		return err;

	switch (cmd)
	{
		case APCI8806_SET_DA_DATA:
			return apci8806_ioc_set_da_data(file, arg);

		case APCI8806_GET_EXIN:
			return apci8806_ioc_get_exin(file, arg);
	}

	return 0;
}

/*--------------------------------------------------------------------------*/
/*  apci8806_open - ǥХΥץ					    */
/*--------------------------------------------------------------------------*/
static int apci8806_open(struct inode* inode, struct file* file)
{
	int ref;
	int result;

	PDEBUG("Open !! %ld\n", file->f_version);

	// Ȥ
	ref = NUM(inode->i_rdev);

	PDEBUG("ref index=%d\n", ref);

	// ǥХξ
	result = get_device_state(ref);
	if (result < 0)
		return result;

	// Ѿ֤ˤ
	devctl[ref].f_version = file->f_version;

	// ǡȰ¸
	file->private_data = (void*)ref;

	MOD_INC_USE_COUNT;		// ȥ+

	return 0;
}

/*--------------------------------------------------------------------------*/
/*  apci8806_close - ǥХΥ					    */
/*--------------------------------------------------------------------------*/
static int apci8806_close(struct inode* inode, struct file* file)
{
	int ref;

	PDEBUG("Close !!\n");

	// ǡλȰ
	ref = (int)file->private_data;

	// ѿʤɤΥ꡼󥢥å׽

	// ̤Ѿ֤ˤ
	devctl[ref].f_version = 0;

	MOD_DEC_USE_COUNT;		// ȥ-

	return 0;
}

/*--------------------------------------------------------------------------*/
/*  init_driver_info - ǥХɥ饤дѿν			    */
/*  Input : none							    */
/*  Output: return - 0 					    */
/*--------------------------------------------------------------------------*/
static int init_driver_info(void)
{
	int i;

	for (i = 0; i < APCI8806_MAX_DEVICES; i++)
	{
		devctl[i].io_addr    = 0x0000;
		devctl[i].f_version  = 0;
	}

	return 0;
}

/*--------------------------------------------------------------------------*/
/*  setup_driver - ǥХɥ饤ФΥåȥå			    */
/*  Input : none							    */
/*  Output: return - 0 ʳ åȥå׼				    */
/*		     0 åȥå					    */
/*--------------------------------------------------------------------------*/
static int setup_driver(void)
{
	int result;

	result = get_device();
	if (result < 0)
		return result;

	// character device driver registration
	result = regist_char_driver();
	if (result < 0)
		return result;

	return 0;
}

/*--------------------------------------------------------------------------*/
/*  release_driver -  ǥХɥ饤Фβ				    */
/*  Input : none							    */
/*  Output: none							    */
/*--------------------------------------------------------------------------*/
static void release_driver(void)
{
	// character device driver unregistration
	unregist_char_driver();
}

/*--------------------------------------------------------------------------*/
/*  get_device - 󥹥ȡ뤵ƤǥХ		    */
/*  Input : none							    */
/*  Output: return - -ENODEV ǥХϸդʤ			    */
/*		     0	     ǥХϸդä			    */
/*--------------------------------------------------------------------------*/
static int get_device(void)
{
	struct pci_dev* pdev = NULL;

	unsigned int  io_addr;		// I/O Base Address

	int i;

	for (i = 0; i < APCI8806_MAX_DEVICES; i++)
	{
		// PCI ǥХ
		pdev = pci_find_device(APCI8806_VENDOR_ID,
				       APCI8806_DEVICE_ID, pdev);
		if (pdev == NULL)
		    break;

		//  I/O ١ɥ쥹
		io_addr	  = get_device_base_addr(pdev->base_address[2]);

		PDEBUG("io_addr:   %X\n", io_addr);

		// ǡ
		devctl[i].io_addr = io_addr;	// I/O Base Address save
	}

	if (i == 0)
	{
		PWARNING("Not found Device(s)\n");
		return -ENODEV;
	}

	return 0;
}

/*--------------------------------------------------------------------------*/
/*  get_device_base_addr - ѲǽʥǥХΥ١ɥ쥹	    */
/*  Input : base_addr - ١ɥ쥹					    */
/*  Output: return - Խ١ɥ쥹				    */
/*--------------------------------------------------------------------------*/
static unsigned int get_device_base_addr(unsigned int base_addr)
{
	// I/O ١ɥ쥹쥸
	if ((base_addr & PCI_BASE_ADDRESS_SPACE) == PCI_BASE_ADDRESS_SPACE_IO)
		return base_addr & PCI_BASE_ADDRESS_IO_MASK;

	// Memory ١ɥ쥹쥸
	return base_addr & PCI_BASE_ADDRESS_MEM_MASK;
}

/*--------------------------------------------------------------------------*/
/*  regist_char_driver - ᥸㡼ֹưŪƤȥɥ饤ФϿ	    */
/*  Input : none							    */
/*  Output: return - register_chrdev 				    */
/*		     0	Ͽ					    */
/*--------------------------------------------------------------------------*/
static int regist_char_driver(void)
{
	int result;

	PDEBUG("devmajor=%d\n", devmajor);

	if (devmajor == 0)
	{
		// ᥸㡼ֹ椬ꤵƤʤ
		// ưŪ᥸㡼ֹμ
		result = register_chrdev(0, APCI8806_NAME, NULL);
		if (result < 0)
		{
			PDEBUG("get major device number faild.\n");
			return result;
		}

		uapci8806major = result;
	}
	else
	{
		// ᥸㡼ֹ椬ꤵƤ
		uapci8806major = devmajor;
	}

	PDEBUG("major number=%u\n", uapci8806major);

	// 饯ǥХɥ饤ФϿ
	result = register_chrdev(uapci8806major, APCI8806_NAME, &apci8806_fops);
	if (result < 0)
	{
		PDEBUG("register_chrdev faild.\n");
		return result;
	}

	return 0;
}

/*--------------------------------------------------------------------------*/
/*  unregist_char_driver - ɥ饤ФϿä			    */
/*  Input : none							    */
/*  Output: return - unregister_chrdev 				    */
/*		     0	Ͽ					    */
/*--------------------------------------------------------------------------*/
static int unregist_char_driver(void)
{
	int result;

	// 饯ǥХɥ饤ФϿ
	result = unregister_chrdev(uapci8806major, APCI8806_NAME);
	if (result < 0)
	{
		PDEBUG("unregister_chrdev faild.\n");
		return result;
	}

	return 0;
}

/*--------------------------------------------------------------------------*/
/*  get_device_state - ǥХξ֤				    */
/*  Input : ref	  - Ȱ						    */
/*  Output: return - -ENODEV ΰ֤˥ǥХϤʤ			    */
/*		     -EBUSY  					    */
/*		     0	     Ѳǽ 					    */
/*  Notes : ref >= 0 - ref <= 15 ޤǤʤ				    */
/*--------------------------------------------------------------------------*/
static int get_device_state(int ref)
{
	if (devctl[ref].io_addr == 0x0000)
		return -ENODEV;

	if (devctl[ref].f_version != 0)
		return -EBUSY;

	return 0;
}

/*--------------------------------------------------------------------------*/
/*  verify_arg - ǥХͭΰ				    */
/*  Input : cmd - ioctl 楳ޥ					    */
/*	    arg - 桼֤ؤΥɥ쥹				    */
/*  Output: return - -EINVAL ̵ʰ					    */
/*		     verify_area 				    */
/*		     0	     Ѳǽ 					    */
/*--------------------------------------------------------------------------*/
static int verify_arg(unsigned int cmd, unsigned long arg)
{
	int size = _IOC_SIZE(cmd);

	/* don't even decode wrong cmds: better returning  EINVAL than EFAULT */
	if (_IOC_TYPE(cmd) != APCI8806_IOC_MAGIC)
		return -EINVAL;

	if (_IOC_NR(cmd) > APCI8806_IOC_MAXNR)
		return -EINVAL;

	/*
	 * the type is a bitmask, and VERIFY_WRITE catches R/W
	 * transfers. Note that the type is user-oriented, while
	 * verify_area is kernel-oriented, so the concept of "read" and
	 * "write" is reversed
	 */
	if (_IOC_TYPE(cmd) & _IOC_READ)
		return verify_area(VERIFY_WRITE, (void*)arg, size);

	if (_IOC_TYPE(cmd) & _IOC_WRITE)
		return verify_area(VERIFY_READ, (void*)arg, size);

	return 0;
}

/*--------------------------------------------------------------------------*/
/*  apci8806_ioc_set_da_data - D/A ǡ				    */
/*  Input : file - ͥ file ¤ΤΥݥ			    */
/*	    arg	 - 桼֤ؤΥɥ쥹				    */
/*  Output: return - -EINVAL    ̵ʰ				    */
/*		     -ETIMEDOUT ॢ				    */
/*		     0	     ｪλ					    */
/*--------------------------------------------------------------------------*/
static int apci8806_ioc_set_da_data(struct file* file, unsigned long arg)
{
	int ref, i;
	apci8806_da data;
	unsigned short da;

	PDEBUG("Set da data\n");

	// ǡλȰ
	ref = (int)file->private_data;

	// 桼֤ǡ
	copy_from_user(&data, (void*)arg, sizeof(data));

	// ͥ
	if (!is_channel(data.ch))
		return -EINVAL;

	// D/A ǡ
	da = (data.value << 4) & 0xFFF0;

	// D/A ͥ
	da |= data.ch;

	// D/A ǡž
	for (i = 0; i < APCI8803_DA_TIMEOUT; i++)
	{
		// ƥơϥǡ
		if (!(da_get_data(ref) & APCI8806_DA_BUSY))
			break;

		udelay(1);	// 1us wait
	}

	if (i >= APCI8803_DA_TIMEOUT)
		return -ETIMEDOUT;

	// D/A ǡ
	da_set_data(ref, da);

	return 0;
}

/*--------------------------------------------------------------------------*/
/*  apci8806_ioc_get_exin - ϥݡȤǡ		    */
/*  Input : file - ͥ file ¤ΤΥݥ			    */
/*	    arg	 - 桼֤ؤΥɥ쥹				    */
/*  Output: return - 0	     ｪλ					    */
/*--------------------------------------------------------------------------*/
static int apci8806_ioc_get_exin(struct file* file, unsigned long arg)
{
	int ref;
	unsigned char data;

	PDEBUG("Get EX IN\n");

	// ǡλȰ
	ref = (int)file->private_data;

	// ƥơϥǡ
	data = da_get_data(ref);

	// ɬפʾΤߤˤ
	data &= (APCI8806_EXIN1 | APCI8806_EXIN2);

	// 桼֤إǡ
	put_user(data, (unsigned char*)arg);

	return 0;
}

/*--------------------------------------------------------------------------*/
/*  is_channel - D/A ͥå					    */
/*  Input : ch	 - ͥ						    */
/*  Output: return - 0 ϰϳ						    */
/*		     1 ϰ						    */
/*--------------------------------------------------------------------------*/
static int is_channel(int ch)
{
	// ץ󥰳ϥͥ(0  15)
	if (ch < APCI8806_CH_FIRST || ch > APCI8806_CH_LAST)
		return 0;

	return 1;
}

/*--------------------------------------------------------------------------*/
/*  da_set_data - D/A ǡ					    */
/*  Input : ref  - Ȱ						    */
/*	    data - ǡ						    */
/*  Output: none							    */
/*--------------------------------------------------------------------------*/
static inline void da_set_data(int ref, unsigned short data)
{
	outw(data, devctl[ref].io_addr);
}

/*--------------------------------------------------------------------------*/
/*  da_get_data - ƥơϥǡ			    */
/*  Input : ref - Ȱ						    */
/*  Output: return - ƥơϥǡ			    */
/*--------------------------------------------------------------------------*/
static inline unsigned char da_get_data(int ref)
{
	return inb(devctl[ref].io_addr);
}
